@@ -99,7 +99,11 @@ translate_events_solver <- function(events, sample, timestamps) {
       tool_result <- turn@contents[[1]]
       events <- c(
         events,
-        create_tool_event(turn, tool_result, timestamps = timestamps)
+        create_tool_event(
+          turn,
+          tool_result,
+          timestamp = timestamps$solve$started_at
+        )
       )
       next
     }
@@ -127,9 +131,48 @@ translate_events_solver <- function(events, sample, timestamps) {
 translate_events_scorer <- function(events, sample, timestamps = timestamps) {
   if ("scorer_chat" %in% names(sample)) {
     scorer_chat <- sample$scorer_chat[[1]]
+    scorer_turns <- scorer_chat$get_turns()
     scorer_turn <- scorer_chat$last_turn()
     time_scorer <- timestamps$score$started_at
 
+    if (has_tool_calls(scorer_turns)) {
+      events <- c(
+        events,
+        create_use_tools_begin_event(
+          time_scorer,
+          attr(scorer_turns[[1]], "working_start") %||% 0,
+          type = "scorer"
+        )
+      )
+      events <- c(events, create_tool_state_event(time_scorer, scorer_chat))
+
+      tool_result_turns <- purrr::keep(scorer_turns, function(turn) {
+        length(turn@contents) == 1 &&
+          inherits(turn@contents[[1]], "ellmer::ContentToolResult")
+      })
+
+      for (turn in tool_result_turns) {
+        tool_result <- turn@contents[[1]]
+        events <- c(
+          events,
+          create_tool_event(
+            turn,
+            tool_result,
+            timestamp = timestamps$score$started_at
+          )
+        )
+      }
+
+      events <- c(
+        events,
+        create_use_tools_end_event(
+          time_scorer,
+          attr(scorer_turns[[length(scorer_turns)]], "working_start") %||% 0,
+          type = "scorer"
+        )
+      )
+    }
+
     events <- c(
       events,
       create_scorer_begin_event(
@@ -137,14 +180,32 @@ translate_events_scorer <- function(events, sample, timestamps = timestamps) {
         attr(scorer_turn, "working_start")
       )
     )
-    events <- c(
-      events,
-      create_scoring_model_event(
-        scorer_turn,
-        sample,
-        time_scorer
-      )
+
+    assistant_turns <- purrr::keep(
+      scorer_turns,
+      function(turn) turn@role == "assistant"
     )
+
+    for (turn in assistant_turns) {
+      has_tool_request <- any(vapply(
+        turn@contents,
+        inherits,
+        logical(1),
+        "ellmer::ContentToolRequest"
+      ))
+
+      if (identical(turn, scorer_turn) || has_tool_request) {
+        events <- c(
+          events,
+          create_scoring_model_event(
+            turn,
+            sample,
+            time_scorer
+          )
+        )
+      }
+    }
+
     events <- c(events, create_score_event(scorer_turn, sample, time_scorer))
     events <- c(
       events,
@@ -171,13 +232,23 @@ create_init_begin_event <- function(timestamp) {
 
 create_sample_init_event <- function(turn, sample, timestamp) {
   user_message_id <- generate_id()
+  message_content <- message_content_from_turn(turn)
+  sample_input <- if (is_content_list(message_content)) {
+    list(list(
+      content = message_content,
+      source = "input",
+      role = "user"
+    ))
+  } else {
+    message_content
+  }
 
   list(list(
     timestamp = events_timestamp(timestamp),
     working_start = attr(turn, "working_start"),
     event = "sample_init",
     sample = list(
-      input = input_string(sample$input[[1]]),
+      input = sample_input,
       target = sample$target,
       id = sample$id
     ),
@@ -185,7 +256,7 @@ create_sample_init_event <- function(turn, sample, timestamp) {
       messages = list(
         list(
           id = user_message_id,
-          content = input_string(sample$input[[1]]),
+          content = message_content,
           source = "input",
           role = "user"
         )
@@ -213,13 +284,17 @@ create_init_end_event <- function(timestamp, working_start) {
   ))
 }
 
-create_use_tools_begin_event <- function(timestamp, working_start) {
+create_use_tools_begin_event <- function(
+  timestamp,
+  working_start,
+  type = "solver"
+) {
   list(list(
     timestamp = events_timestamp(timestamp),
     working_start = working_start,
     event = "step",
     action = "begin",
-    type = "solver",
+    type = type,
     name = "use_tools"
   ))
 }
@@ -270,19 +345,22 @@ create_tool_state_event <- function(timestamp, chat) {
   ))
 }
 
-create_use_tools_end_event <- function(timestamp, working_start) {
+create_use_tools_end_event <- function(
+  timestamp,
+  working_start,
+  type = "solver"
+) {
   list(list(
     timestamp = events_timestamp(timestamp),
     working_start = working_start,
     event = "step",
     action = "end",
-    type = "solver",
+    type = type,
     name = "use_tools"
   ))
 }
 
-create_tool_event <- function(turn, tool_result, timestamps) {
-  timestamp <- timestamps$solve$started_at
+create_tool_event <- function(turn, tool_result, timestamp) {
   list(list(
     timestamp = events_timestamp(timestamp),
     working_start = attr(turn, "working_start"),
@@ -291,7 +369,11 @@ create_tool_event <- function(turn, tool_result, timestamps) {
     id = tool_result@request@id,
     `function` = tool_result@request@name,
     arguments = tool_result@request@arguments,
-    result = tool_result@value %||% as.character(tool_result@error),
+    result = if (!is.null(tool_result@error)) {
+      as.character(tool_result@error)
+    } else {
+      collapse_tool_result(tool_result)
+    },
     events = list(),
     completed = events_timestamp(timestamp),
     working_time = attr(turn, "working_time")
@@ -332,7 +414,11 @@ create_model_event <- function(turn, sample) {
         tool_result <- prev_turn@contents[[1]]
         return(list(
           id = generate_id(),
-          content = tool_result@value %||% as.character(tool_result@error),
+          content = if (!is.null(tool_result@error)) {
+            as.character(tool_result@error)
+          } else {
+            collapse_tool_result(tool_result)
+          },
           role = "tool",
           tool_call_id = tool_result@request@id,
           `function` = tool_result@request@name
@@ -340,7 +426,7 @@ create_model_event <- function(turn, sample) {
       } else {
         return(list(
           id = generate_id(),
-          content = prev_turn@text,
+          content = message_content_from_turn(prev_turn),
           source = "input",
           role = "user"
         ))
@@ -394,25 +480,7 @@ create_model_event <- function(turn, sample) {
 
   stop_reason <- ifelse(has_tool_calls_in_turn, "tool_calls", "stop")
 
-  tools_list <- list()
-  if (length(solver_chat$get_tools()) > 0) {
-    tools <- solver_chat$get_tools()
-    tools_list <- lapply(seq_along(tools), function(i) {
-      tool <- tools[[i]]
-      tool_name <- names(tools)[i]
-
-      list(
-        name = tool_name,
-        description = tool@description,
-        parameters = list(
-          type = "object",
-          properties = c(),
-          required = list(),
-          additionalProperties = FALSE
-        )
-      )
-    })
-  }
+  tools_list <- chat_tools_list(solver_chat)
 
   output_message <- list(
     id = generate_id(),
@@ -434,16 +502,44 @@ create_model_event <- function(turn, sample) {
         content = list(list(
           tool_use_id = msg$tool_call_id,
           type = "tool_result",
-          content = list(list(type = "text", text = msg$content)),
+          content = if (is.character(msg$content)) {
+            list(list(type = "text", text = msg$content))
+          } else {
+            # Handle tool results that may contain image objects
+            lapply(msg$content, function(item) {
+              if (
+                is.list(item) &&
+                  identical(item$type, "image") &&
+                  !is.null(item$source)
+              ) {
+                # Convert image object to ContentImage format
+                list(
+                  type = "image",
+                  image = paste0(
+                    "data:",
+                    item$source$media_type,
+                    ";base64,",
+                    item$source$data
+                  )
+                )
+              } else {
+                item
+              }
+            })
+          },
           # This depends specifically on previous helpers using
           # `as_character()` on conditions to extract error messages
-          is_error = grepl("Error in", msg$content)
+          is_error = if (is.character(msg$content)) {
+            grepl("Error in", msg$content)
+          } else {
+            FALSE
+          }
         ))
       ))
     } else if (msg$role == "user") {
       return(list(
         role = "user",
-        content = list(list(type = "text", text = msg$content))
+        content = ensure_content_list(msg$content)
       ))
     } else if (msg$role == "assistant") {
       if ("tool_calls" %in% names(msg)) {
@@ -466,9 +562,35 @@ create_model_event <- function(turn, sample) {
           content = combined_content
         ))
       } else {
+        # Handle content that may contain image objects
+        processed_content <- if (is.list(msg$content)) {
+          lapply(msg$content, function(item) {
+            if (
+              is.list(item) &&
+                identical(item$type, "image") &&
+                !is.null(item$source)
+            ) {
+              # Convert image object to ContentImage format
+              list(
+                type = "image",
+                image = paste0(
+                  "data:",
+                  item$source$media_type,
+                  ";base64,",
+                  item$source$data
+                )
+              )
+            } else {
+              item
+            }
+          })
+        } else {
+          msg$content
+        }
+
         return(list(
           role = "assistant",
-          content = msg$content
+          content = processed_content
         ))
       }
     }
@@ -500,8 +622,11 @@ create_model_event <- function(turn, sample) {
       request = list(
         messages = request_messages,
         tools = tools_list,
-        tool_choice = if (length(tools_list) > 0) list(type = "auto") else
-          "none",
+        tool_choice = if (length(tools_list) > 0) {
+          list(type = "auto")
+        } else {
+          "none"
+        },
         model = solver_chat$get_model(),
         max_tokens = 4096,
         extra_headers = list(
@@ -547,6 +672,29 @@ create_model_event <- function(turn, sample) {
   ))
 }
 
+chat_tools_list <- function(chat) {
+  if (length(chat$get_tools()) == 0) {
+    return(list())
+  }
+
+  tools <- chat$get_tools()
+  lapply(seq_along(tools), function(i) {
+    tool <- tools[[i]]
+    tool_name <- names(tools)[i]
+
+    list(
+      name = tool_name,
+      description = tool@description,
+      parameters = list(
+        type = "object",
+        properties = c(),
+        required = list(),
+        additionalProperties = FALSE
+      )
+    )
+  })
+}
+
 create_solver_end_event <- function(timestamp, working_start) {
   list(list(
     timestamp = events_timestamp(timestamp),
@@ -570,23 +718,209 @@ create_scorer_begin_event <- function(timestamp, working_start) {
 }
 
 create_scoring_model_event <- function(turn, sample, timestamp) {
-  user_id <- generate_id()
   scorer_chat <- sample$scorer_chat[[1]]
+  turns <- scorer_chat$get_turns()
+  previous_turns <- list()
+
+  for (j in seq_along(turns)) {
+    if (identical(turns[[j]], turn)) {
+      break
+    }
+    previous_turns[[length(previous_turns) + 1]] <- turns[[j]]
+  }
+
+  input_messages <- lapply(previous_turns, function(prev_turn) {
+    if (prev_turn@role == "user") {
+      if (
+        length(prev_turn@contents) == 1 &&
+          inherits(prev_turn@contents[[1]], "ellmer::ContentToolResult")
+      ) {
+        tool_result <- prev_turn@contents[[1]]
+        return(list(
+          id = generate_id(),
+          content = if (!is.null(tool_result@error)) {
+            as.character(tool_result@error)
+          } else {
+            collapse_tool_result(tool_result)
+          },
+          role = "tool",
+          tool_call_id = tool_result@request@id,
+          `function` = tool_result@request@name
+        ))
+      } else {
+        return(list(
+          id = generate_id(),
+          content = message_content_from_turn(prev_turn),
+          source = "input",
+          role = "user"
+        ))
+      }
+    } else {
+      message <- list(
+        id = generate_id(),
+        content = list(list(type = "text", text = prev_turn@text)),
+        source = "generate",
+        role = "assistant"
+      )
+
+      tool_requests <- purrr::keep(prev_turn@contents, function(content) {
+        inherits(content, "ellmer::ContentToolRequest")
+      })
+
+      if (length(tool_requests) > 0) {
+        tool_calls <- lapply(tool_requests, function(req) {
+          list(
+            id = req@id,
+            `function` = req@name,
+            arguments = req@arguments
+          )
+        })
+
+        message$tool_calls <- tool_calls
+      }
+
+      return(message)
+    }
+  })
+
+  has_tool_calls_in_turn <- any(sapply(turn@contents, function(content) {
+    inherits(content, "ellmer::ContentToolRequest")
+  }))
+
+  tool_calls_list <- list()
+  if (has_tool_calls_in_turn) {
+    tool_requests <- purrr::keep(turn@contents, function(content) {
+      inherits(content, "ellmer::ContentToolRequest")
+    })
+
+    tool_calls_list <- lapply(tool_requests, function(req) {
+      list(
+        id = req@id,
+        `function` = req@name,
+        arguments = req@arguments
+      )
+    })
+  }
+
+  stop_reason <- ifelse(has_tool_calls_in_turn, "tool_calls", "stop")
+  tools_list <- chat_tools_list(scorer_chat)
+
+  output_message <- list(
+    id = generate_id(),
+    content = list(list(type = "text", text = turn@text)),
+    source = "generate",
+    role = "assistant"
+  )
+
+  if (has_tool_calls_in_turn) {
+    output_message$tool_calls <- tool_calls_list
+  }
+
+  output_message$model <- scorer_chat$get_model()
+
+  request_messages <- lapply(input_messages, function(msg) {
+    if (msg$role == "tool") {
+      return(list(
+        role = "user",
+        content = list(list(
+          tool_use_id = msg$tool_call_id,
+          type = "tool_result",
+          content = if (is.character(msg$content)) {
+            list(list(type = "text", text = msg$content))
+          } else {
+            lapply(msg$content, function(item) {
+              if (
+                is.list(item) &&
+                  identical(item$type, "image") &&
+                  !is.null(item$source)
+              ) {
+                list(
+                  type = "image",
+                  image = paste0(
+                    "data:",
+                    item$source$media_type,
+                    ";base64,",
+                    item$source$data
+                  )
+                )
+              } else {
+                item
+              }
+            })
+          },
+          is_error = if (is.character(msg$content)) {
+            grepl("Error in", msg$content)
+          } else {
+            FALSE
+          }
+        ))
+      ))
+    } else if (msg$role == "user") {
+      return(list(
+        role = "user",
+        content = ensure_content_list(msg$content)
+      ))
+    } else if (msg$role == "assistant") {
+      if ("tool_calls" %in% names(msg)) {
+        tool_use_elements <- lapply(msg$tool_calls, function(tc) {
+          list(
+            type = "tool_use",
+            id = tc$id,
+            name = tc$`function`,
+            input = tc$arguments
+          )
+        })
+
+        combined_content <- c(
+          list(list(type = "text", text = msg$content[[1]]$text)),
+          tool_use_elements
+        )
+
+        return(list(
+          role = "assistant",
+          content = combined_content
+        ))
+      } else {
+        processed_content <- if (is.list(msg$content)) {
+          lapply(msg$content, function(item) {
+            if (
+              is.list(item) &&
+                identical(item$type, "image") &&
+                !is.null(item$source)
+            ) {
+              list(
+                type = "image",
+                image = paste0(
+                  "data:",
+                  item$source$media_type,
+                  ";base64,",
+                  item$source$data
+                )
+              )
+            } else {
+              item
+            }
+          })
+        } else {
+          msg$content
+        }
+
+        return(list(
+          role = "assistant",
+          content = processed_content
+        ))
+      }
+    }
+  })
 
   list(list(
     timestamp = events_timestamp(timestamp),
     working_start = attr(turn, "working_start"),
     event = "model",
     model = scorer_chat$get_model(),
-    input = list(
-      list(
-        id = user_id,
-        content = turn@text,
-        role = "user"
-      )
-    ),
-    tools = list(),
-    tool_choice = "none",
+    input = input_messages,
+    tools = tools_list,
+    tool_choice = if (length(tools_list) > 0) "auto" else "none",
     config = list(
       max_tokens = 4096
     ),
@@ -594,18 +928,8 @@ create_scoring_model_event <- function(turn, sample, timestamp) {
       model = scorer_chat$get_model(),
       choices = list(
         list(
-          message = list(
-            id = generate_id(),
-            content = list(
-              list(
-                type = "text",
-                text = turn@text
-              )
-            ),
-            source = "generate",
-            role = "assistant"
-          ),
-          stop_reason = "stop"
+          message = output_message,
+          stop_reason = stop_reason
         )
       ),
       usage = turn_tokens(turn),
@@ -613,13 +937,13 @@ create_scoring_model_event <- function(turn, sample, timestamp) {
     ),
     call = list(
       request = list(
-        messages = list(
-          list(
-            role = "user",
-            content = turn@text
-          )
-        ),
-        tools = list(),
+        messages = request_messages,
+        tools = tools_list,
+        tool_choice = if (length(tools_list) > 0) {
+          list(type = "auto")
+        } else {
+          "none"
+        },
         model = scorer_chat$get_model(),
         max_tokens = 4096,
         extra_headers = list(
@@ -628,16 +952,32 @@ create_scoring_model_event <- function(turn, sample, timestamp) {
       ),
       response = list(
         id = paste0("msg_", generate_id()),
-        content = list(
-          list(
+        content = if (has_tool_calls_in_turn) {
+          c(
+            list(list(
+              citations = NULL,
+              text = turn@text,
+              type = "text"
+            )),
+            lapply(tool_calls_list, function(tc) {
+              list(
+                id = tc$id,
+                input = tc$arguments,
+                name = tc$`function`,
+                type = "tool_use"
+              )
+            })
+          )
+        } else {
+          list(list(
             citations = NULL,
             text = turn@text,
             type = "text"
-          )
-        ),
+          ))
+        },
         model = scorer_chat$get_model(),
         role = "assistant",
-        stop_reason = "end_turn",
+        stop_reason = if (has_tool_calls_in_turn) "tool_use" else "end_turn",
         stop_sequence = NULL,
         type = "message",
         usage = turn_tokens(turn),
@@ -650,19 +990,21 @@ create_scoring_model_event <- function(turn, sample, timestamp) {
 }
 
 create_score_event <- function(turn, sample, timestamp) {
+  scorer_user_turn <- sample$scorer_chat[[1]]$get_turns()[[1]]
+
   list(list(
     timestamp = events_timestamp(timestamp),
     working_start = attr(turn, "working_start"),
     event = "score",
     score = list(
-      value = "C",
-      answer = turn@text,
+      value = sample$score,
+      answer = as.character(sample$result),
       explanation = turn@text,
       metadata = list(
         grading = list(
           list(
             id = generate_id(),
-            content = turn@text,
+            content = scorer_user_turn@text,
             role = "user"
           ),
           list(
@@ -717,16 +1059,19 @@ turn_tokens <- function(turn) {
   )
 }
 
-# log working_time and working_starts values by pre-computing them from the Chat
-# objects before mapping over turns (#97).
-# `working_start` is the clock time when a turn started minus the clock time
-# when the first chat in the solver or scorer started, in seconds.
-# `working_time` is the duration of the turn (completed time of the turn
-# minus the completed time of the turn preceding it).
-#
-# this was initially added to process `@completed` slots of turns. those were
-# removed in ellmer 0.2.0, so we're temporarily filling them in with
-# the average timing (#112)
+turn_duration <- function(turn) {
+  duration <- 0
+  if (inherits(turn, "ellmer::AssistantTurn")) {
+    if (!is.na(turn@duration)) {
+      duration <- turn@duration
+    }
+  }
+  duration
+}
+
+# log working_time values by extracting them from the Turn @duration slots (#115).
+# `working_time` is the duration of the turn in seconds. User turns have NA duration,
+# assistant turns have the actual request duration as measured by httr2.
 add_working_times_to_turns <- function(chat, which, timestamps, n) {
   turns <- chat$get_turns()
 
@@ -734,18 +1079,8 @@ add_working_times_to_turns <- function(chat, which, timestamps, n) {
     return(chat)
   }
 
-  average_working_time <-
-    as.numeric(difftime(
-      timestamps[[which]]$completed_at,
-      timestamps[[which]]$started_at,
-      units = "secs"
-    )) /
-    (length(turns) * n)
-
-  attr(turns[[1]], "working_time") <- NA_real_
-  for (i in 2:length(turns)) {
-    # TODO: revisit once durations are added to ellmer turns (#112)
-    attr(turns[[i]], "working_time") <- average_working_time
+  for (i in seq_along(turns)) {
+    attr(turns[[i]], "working_time") <- turn_duration(turns[[i]])
   }
 
   chat$set_turns(turns)
@@ -753,27 +1088,22 @@ add_working_times_to_turns <- function(chat, which, timestamps, n) {
   chat
 }
 
-# this was initially added to process `@completed` slots of turns. those were
-# removed in ellmer 0.2.0, so we're temporarily filling them in with
-# the estimated timing if every sample took the same amount of time (#112)
+# log working_start values by accumulating turn durations (#115).
+# `working_start` is the clock time when a turn started minus the clock time
+# when the first chat started, in seconds. This is computed by accumulating
+# the @duration values from previous turns.
 add_working_start_to_turns <- function(chats, which, timestamps) {
-  total_duration <-
-    as.numeric(difftime(
-      timestamps[[which]]$completed_at,
-      timestamps[[which]]$started_at,
-      units = "secs"
-    ))
-  duration_per_chat <- total_duration / length(chats)
   current_working_start <- 0
 
-  for (i in 1:length(chats)) {
+  for (i in seq_along(chats)) {
     chat <- chats[[i]]
     chat_turns <- chat$get_turns()
-    duration_per_turn <- duration_per_chat / (length(chat_turns) - 1)
-    for (j in 1:length(chat_turns)) {
+
+    for (j in seq_along(chat_turns)) {
       attr(chat_turns[[j]], "working_start") <- current_working_start
-      current_working_start <- current_working_start + duration_per_turn
+      current_working_start <- current_working_start + turn_duration(chat_turns[[j]])
     }
+
     chat$set_turns(chat_turns)
     chats[[i]] <- chat
   }
